@@ -4,6 +4,11 @@
 
 #include <time.h>
 #include <string.h>
+#include <dlfcn.h>
+
+#ifndef RTLD_DEFAULT
+#define RTLD_DEFAULT ((void*)0)
+#endif
 
 /* Load STDC format macros (PRI*), or define them
  * for those crappy, non-standard compilers
@@ -30,8 +35,22 @@ struct platform_data {
 	cl_bool has_amd_offline; /* has cl_amd_offline_devices extension */
 };
 
+struct platform_info_checks {
+	int has_khr_icd;
+	cl_uint plat_version;
+};
+
 cl_uint num_platforms;
 cl_platform_id *platform;
+/* highest version exposed by any platform: if the OpenCL library (the ICD loader)
+ * has a lower version, problems may arise (such as API calls causing segfaults
+ * or any other unexpected behavior
+ */
+cl_uint max_plat_version;
+/* auto-detected OpenCL version support for the ICD loader */
+cl_uint icdl_ocl_version_found = 10;
+/* OpenCL version support declared by the ICD loader */
+cl_uint icdl_ocl_version;
 
 struct platform_data *pdata;
 /* maximum length of a platform's sname */
@@ -109,7 +128,7 @@ static const char l3cache[] = "L3 cache";
 static const char l4cache[] = "L4 cache";
 
 static const char* affinity_domain_str[] = {
-	numa, l4cache, l3cache, l2cache, l1cache, "next partitionalbe"
+	numa, l4cache, l3cache, l2cache, l1cache, "next partitionable"
 };
 
 static const char* affinity_domain_ext_str[] = {
@@ -208,21 +227,21 @@ static const char* sources[] = {
 	"KRN()\n/* KRN(2)\nKRN(4)\nKRN(8)\nKRN(16) */\n",
 };
 
-const char *no_plat()
+const char *no_plat(void)
 {
 	return output_mode == CLINFO_HUMAN ?
 		"No platform" :
 		"CL_INVALID_PLATFORM";
 }
 
-const char *no_dev()
+const char *no_dev(void)
 {
 	return output_mode == CLINFO_HUMAN ?
 		"No devices found in platform" :
 		"CL_DEVICE_NOT_FOUND";
 }
 
-const char *no_dev_avail()
+const char *no_dev_avail(void)
 {
 	return output_mode == CLINFO_HUMAN ?
 		"No devices available in platform" :
@@ -251,6 +270,30 @@ static const char vbar_str[] = " | ";
 int had_error = 0;
 const char *cur_sfx = empty_str;
 
+/* parse a CL_DEVICE_VERSION or CL_PLATFORM_VERSION info to determine the OpenCL version.
+ * Returns an unsigned integer in the form major*10 + minor
+ */
+cl_uint
+getOpenCLVersion(const char *version)
+{
+	cl_uint ret = 10;
+	long parse = 0;
+	const char *from = version;
+	char *next = NULL;
+	parse = strtol(from, &next, 10);
+
+	if (next != from) {
+		ret = parse*10;
+		// skip the dot TODO should we actually check for the dot?
+		from = ++next;
+		parse = strtol(from, &next, 10);
+		if (next != from)
+			ret += parse;
+	}
+	return ret;
+}
+
+
 /* print strbuf, prefixed by pname, skipping leading whitespace if skip is nonzero,
  * affixing cur_sfx */
 static inline
@@ -263,7 +306,7 @@ void show_strbuf(const char *pname, int skip)
 }
 
 int
-platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname)
+platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks * chk UNUSED)
 {
 	error = clGetPlatformInfo(pid, param, 0, NULL, &nusz);
 	if (nusz > bufsz) {
@@ -272,7 +315,7 @@ platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname)
 	}
 	had_error = REPORT_ERROR2("get %s size");
 	if (!had_error) {
-		error = clGetPlatformInfo(pid, param, bufsz, strbuf, 0);
+		error = clGetPlatformInfo(pid, param, bufsz, strbuf, NULL);
 		had_error = REPORT_ERROR2("get %s");
 	}
 	/* when only listing, do not print anything we're just gathering
@@ -283,14 +326,32 @@ platform_info_str(cl_platform_id pid, cl_platform_info param, const char* pname)
 	return had_error;
 }
 
-struct platform_info_checks {
-	int has_khr_icd;
-};
+int
+platform_info_ulong(cl_platform_id pid, cl_platform_info param, const char* pname, const struct platform_info_checks * chk UNUSED)
+{
+	cl_ulong val = 0;
+
+	error = clGetPlatformInfo(pid, param, sizeof(val), &val, NULL);
+	had_error = REPORT_ERROR2("get %s");
+	/* when only listing, do not print anything we're just gathering
+	 * information
+	 */
+	if (!list_only) {
+		if (had_error)
+			show_strbuf(pname, 0);
+		else
+			printf("%s" I1_STR "%" PRIu64 "%s\n", line_pfx, pname, val, cur_sfx);
+	}
+	return had_error;
+}
 
 struct platform_info_traits {
 	cl_platform_info param; // CL_PLATFORM_*
 	const char *sname; // "CL_PLATFORM_*"
 	const char *pname; // "Platform *"
+	const char *sfx; // suffix for the output in non-raw mode
+	/* pointer to function that shows the parameter */
+	int (*show_func)(cl_platform_id pid, cl_platform_info param, const char *pname, const struct platform_info_checks *);
 	/* pointer to function that checks if the parameter should be checked */
 	int (*check_func)(const struct platform_info_checks *);
 };
@@ -300,15 +361,21 @@ int khr_icd_p(const struct platform_info_checks *chk)
 	return chk->has_khr_icd;
 }
 
-#define PINFO_COND(symbol, name, funcptr) { symbol, #symbol, "Platform " name, &funcptr }
-#define PINFO(symbol, name) { symbol, #symbol, "Platform " name, NULL }
+int plat_is_21(const struct platform_info_checks *chk)
+{
+	return !(chk->plat_version < 21);
+}
+
+#define PINFO_COND(symbol, name, sfx, typ, funcptr) { symbol, #symbol, "Platform " name, sfx, &platform_info_##typ, &funcptr }
+#define PINFO(symbol, name, sfx, typ) { symbol, #symbol, "Platform " name, sfx, &platform_info_##typ, NULL }
 struct platform_info_traits pinfo_traits[] = {
-	PINFO(CL_PLATFORM_NAME, "Name"),
-	PINFO(CL_PLATFORM_VENDOR, "Vendor"),
-	PINFO(CL_PLATFORM_VERSION, "Version"),
-	PINFO(CL_PLATFORM_PROFILE, "Profile"),
-	PINFO(CL_PLATFORM_EXTENSIONS, "Extensions"),
-	PINFO_COND(CL_PLATFORM_ICD_SUFFIX_KHR, "Extensions function suffix", khr_icd_p)
+	PINFO(CL_PLATFORM_NAME, "Name", NULL, str),
+	PINFO(CL_PLATFORM_VENDOR, "Vendor", NULL, str),
+	PINFO(CL_PLATFORM_VERSION, "Version", NULL, str),
+	PINFO(CL_PLATFORM_PROFILE, "Profile", NULL, str),
+	PINFO(CL_PLATFORM_EXTENSIONS, "Extensions", NULL, str),
+	PINFO_COND(CL_PLATFORM_HOST_TIMER_RESOLUTION, "Host timer resolution", "ns", ulong, plat_is_21),
+	PINFO_COND(CL_PLATFORM_ICD_SUFFIX_KHR, "Extensions function suffix", NULL, str, khr_icd_p)
 };
 
 /* Print platform info and prepare arrays for device info */
@@ -318,20 +385,24 @@ printPlatformInfo(cl_uint p)
 	cl_platform_id pid = platform[p];
 	size_t len = 0;
 
-	struct platform_info_checks pinfo_checks = { 0 };
+	struct platform_info_checks pinfo_checks = { 0, 10 };
 
 	current_function = __func__;
 
 	for (current_line = 0; current_line < ARRAY_SIZE(pinfo_traits); ++current_line) {
 		const struct platform_info_traits *traits = pinfo_traits + current_line;
+		const char *pname = (output_mode == CLINFO_HUMAN ?
+			traits->pname : traits->sname);
+
 		current_param = traits->sname;
 
 		if (traits->check_func && !traits->check_func(&pinfo_checks))
 			continue;
 
-		had_error = platform_info_str(pid, traits->param,
-			output_mode == CLINFO_HUMAN ?
-			traits->pname : traits->sname);
+		cur_sfx = (output_mode == CLINFO_HUMAN && traits->sfx) ? traits->sfx : empty_str;
+
+		had_error = traits->show_func(pid, traits->param,
+			pname, &pinfo_checks);
 
 		if (had_error)
 			continue;
@@ -348,6 +419,10 @@ printPlatformInfo(cl_uint p)
 			memcpy(pdata[p].pname, strbuf, len);
 			pdata[p].pname[len] = '\0';
 			break;
+		case CL_PLATFORM_VERSION:
+			/* compute numeric value for OpenCL version */
+			pinfo_checks.plat_version = getOpenCLVersion(strbuf + 7);
+			break;
 		case CL_PLATFORM_EXTENSIONS:
 			pinfo_checks.has_khr_icd = !!strstr(strbuf, "cl_khr_icd");
 			pdata[p].has_amd_offline = !!strstr(strbuf, "cl_amd_offline_devices");
@@ -367,10 +442,14 @@ printPlatformInfo(cl_uint p)
 
 	}
 
+	if (pinfo_checks.plat_version > max_plat_version)
+		max_plat_version = pinfo_checks.plat_version;
+
 	/* if no CL_PLATFORM_ICD_SUFFIX_KHR, use P### as short/symbolic name */
 	if (!pdata[p].sname) {
-		ALLOC(pdata[p].sname, 32, "platform symbolic name");
-		sprintf(pdata[p].sname, "P%u", p);
+#define SNAME_MAX 32
+		ALLOC(pdata[p].sname, SNAME_MAX, "platform symbolic name");
+		snprintf(pdata[p].sname, SNAME_MAX, "P%u", p);
 	}
 
 	len = strlen(pdata[p].sname);
@@ -407,9 +486,9 @@ getWGsizes(cl_platform_id pid, cl_device_id dev)
 		CL_CONTEXT_PLATFORM, (cl_context_properties)pid,
 		0, 0 };
 	cl_uint cursor = 0;
-	cl_context ctx = 0;
-	cl_program prg = 0;
-	cl_kernel krn = 0;
+	cl_context ctx = NULL;
+	cl_program prg = NULL;
+	cl_kernel krn = NULL;
 
 	ctx = clCreateContext(ctxpft, 1, &dev, NULL, NULL, &error);
 	RR_ERROR("create context");
@@ -441,7 +520,7 @@ getWGsizes(cl_platform_id pid, cl_device_id dev)
 			sizeof(*wgm), wgm + cursor, NULL);
 		RR_ERROR("get kernel info");
 		clReleaseKernel(krn);
-		krn = 0;
+		krn = NULL;
 	}
 
 out:
@@ -454,29 +533,6 @@ out:
 	return ret;
 }
 
-/* parse a CL_DEVICE_VERSION info to determine the OpenCL version.
- * Returns an unsigned integer in the from major*10 + minor
- */
-cl_uint
-getOpenCLVersion(const char *version)
-{
-	cl_uint ret = 10;
-	long parse = 0;
-	const char *from = version;
-	char *next = NULL;
-	parse = strtol(from, &next, 10);
-
-	if (next != from) {
-		ret = parse*10;
-		// skip the dot TODO should we actually check for the dot?
-		from = ++next;
-		parse = strtol(from, &next, 10);
-		if (next != from)
-			ret += parse;
-	}
-	return ret;
-}
-
 /*
  * Device properties/extensions used in traits checks, and relevant functions
  */
@@ -539,6 +595,12 @@ int dev_is_20(const struct device_info_checks *chk)
 	return !(chk->dev_version < 20);
 }
 
+// device supports 2.1
+int dev_is_21(const struct device_info_checks *chk)
+{
+	return !(chk->dev_version < 21);
+}
+
 // device does not support 2.0
 int dev_not_20(const struct device_info_checks *chk)
 {
@@ -635,7 +697,7 @@ void identify_device_extensions(const char *extensions, struct device_info_check
  */
 
 #define _GET_VAL \
-	error = clGetDeviceInfo(dev, param, sizeof(val), &val, 0); \
+	error = clGetDeviceInfo(dev, param, sizeof(val), &val, NULL); \
 	had_error = REPORT_ERROR2("get %s");
 
 #define _GET_VAL_ARRAY \
@@ -692,7 +754,7 @@ int device_info_str_get(cl_device_id dev, cl_device_info param, const char *pnam
 	}
 	had_error = REPORT_ERROR2("get %s size");
 	if (!had_error) {
-		error = clGetDeviceInfo(dev, param, bufsz, strbuf, 0);
+		error = clGetDeviceInfo(dev, param, bufsz, strbuf, NULL);
 		had_error = REPORT_ERROR2("get %s");
 	}
 	return had_error;
@@ -814,13 +876,13 @@ int device_info_time_offset(cl_device_id dev, cl_device_info param, const char *
 	cl_ulong val = 0;
 	GET_VAL;
 	if (!had_error) {
-		int szval = 0;
+		size_t szval = 0;
 		time_t time = val/UINT64_C(1000000000);
-		szval += sprintf(strbuf, "%" PRIu64 "ns (", val);
-		strcpy(strbuf + szval, ctime(&time));
-		szval = strlen(strbuf) - 1;
-		strbuf[szval++] = ')';
-		strbuf[szval++] = '\0';
+		szval += snprintf(strbuf, bufsz, "%" PRIu64 "ns (", val);
+		szval += bufcpy(szval, ctime(&time));
+		/* overwrite ctime's newline with the closing parenthesis */
+		if (szval < bufsz)
+			strbuf[szval - 1] = ')';
 	}
 	show_strbuf(pname, 0);
 	return had_error;
@@ -833,15 +895,11 @@ int device_info_szptr(cl_device_id dev, cl_device_info param, const char *pname,
 	size_t szval = 0, numval = 0;
 	GET_VAL_ARRAY;
 	if (!had_error) {
-		const char *sep = (output_mode == CLINFO_HUMAN ? times_str : spc_str);
-		size_t sepsz = 1;
 		size_t counter = 0;
+		set_separator(output_mode == CLINFO_HUMAN ? times_str : spc_str);
 		szval = 0;
 		for (counter = 0; counter < numval; ++counter) {
-			if (szval > 0) {
-				strcpy(strbuf + szval, sep);
-				szval += sepsz;
-			}
+			add_separator(&szval);
 			szval += snprintf(strbuf + szval, bufsz - szval - 1, "%" PRIuS, val[counter]);
 			if (szval >= bufsz) {
 				trunc_strbuf();
@@ -857,7 +915,7 @@ int device_info_szptr(cl_device_id dev, cl_device_info param, const char *pname,
 int device_info_wg(cl_device_id dev, cl_device_info param UNUSED, const char *pname,
 	const struct device_info_checks *chk UNUSED)
 {
-	cl_platform_id val = 0;
+	cl_platform_id val = NULL;
 	{
 		/* shadow */
 		cl_device_info param = CL_DEVICE_PLATFORM;
@@ -927,30 +985,24 @@ int device_info_devtype(cl_device_id dev, cl_device_info param, const char *pnam
 	GET_VAL;
 	if (!had_error) {
 		/* iterate over device type strings, appending their textual form
-		 * to strbuf. We use plain strcpy since we know that it's at least
-		 * 1024 so we are safe.
+		 * to strbuf.
 		 * TODO: check for extra bits/no bits
 		 */
-		size_t szval = 0;
 		cl_uint i = devtype_count - 1; /* skip CL_DEVICE_TYPE_ALL */
-		const char *sep = (output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
-		size_t sepsz = (output_mode == CLINFO_HUMAN ? 2 : 3);
 		const char * const *devstr = (output_mode == CLINFO_HUMAN ?
 			device_type_str : device_type_raw_str);
+		size_t szval = 0;
+		strbuf[szval] = '\0';
+		set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
 		for (; i > 0; --i) {
 			/* assemble CL_DEVICE_TYPE_* from index i */
 			cl_device_type cur = (cl_device_type)(1) << (i-1);
 			if (val & cur) {
 				/* match: add separator if not first match */
-				if (szval > 0) {
-					strcpy(strbuf + szval, sep);
-					szval += sepsz;
-				}
-				strcpy(strbuf + szval, devstr[i]);
-				szval += strlen(devstr[i]);
+				add_separator(&szval);
+				szval += bufcpy(szval, devstr[i]);
 			}
 		}
-		strbuf[szval] = '\0';
 	}
 	show_strbuf(pname, 0);
 	/* we abuse global strbuf to pass the device type over to the caller */
@@ -1123,12 +1175,11 @@ int device_info_partition_types(cl_device_id dev, cl_device_info param, const ch
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property *val = NULL;
-
-	const char *sep = (output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
-	size_t sepsz = (output_mode == CLINFO_HUMAN ? 2 : 3);
 	const char * const *ptstr = (output_mode == CLINFO_HUMAN ?
 		partition_type_str : partition_type_raw_str);
 
+	set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
+
 	GET_VAL_ARRAY;
 
 	szval = 0;
@@ -1137,10 +1188,7 @@ int device_info_partition_types(cl_device_id dev, cl_device_info param, const ch
 			int str_idx = -1;
 
 			/* add separator for values past the first */
-			if (szval > 0) {
-				strcpy(strbuf + szval, sep);
-				szval += sepsz;
-			}
+			add_separator(&szval);
 
 			switch (val[cursor]) {
 			case 0: str_idx = 1; break;
@@ -1149,7 +1197,7 @@ int device_info_partition_types(cl_device_id dev, cl_device_info param, const ch
 			case CL_DEVICE_PARTITION_BY_AFFINITY_DOMAIN: str_idx = 4; break;
 			case CL_DEVICE_PARTITION_BY_NAMES_INTEL: str_idx = 5; break;
 			default:
-				szval += sprintf(strbuf + szval, "by <unknown> (0x%" PRIXPTR ")", val[cursor]);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "by <unknown> (0x%" PRIXPTR ")", val[cursor]);
 				break;
 			}
 			if (str_idx > 0) {
@@ -1157,16 +1205,17 @@ int device_info_partition_types(cl_device_id dev, cl_device_info param, const ch
 				slen = strlen(ptstr[str_idx]);
 				if (output_mode == CLINFO_RAW && str_idx > 1)
 					slen -= 4;
-				strncpy(strbuf + szval, ptstr[str_idx], slen);
-				szval += slen;
+				szval += bufcpy_len(szval, ptstr[str_idx], slen);
+			}
+			if (szval >= bufsz) {
+				trunc_strbuf();
+				break;
 			}
 		}
 		if (szval == 0) {
-			slen = strlen(ptstr[0]);
-			strcpy(strbuf, ptstr[0]);
-			szval += slen;
-		}
-		strbuf[szval] = '\0';
+			bufcpy(szval, ptstr[0]);
+		} else if (szval < bufsz)
+			strbuf[szval] = '\0';
 	}
 
 	show_strbuf(pname, 0);
@@ -1180,12 +1229,11 @@ int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, cons
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property_ext *val = NULL;
-
-	const char *sep = (output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
-	size_t sepsz = (output_mode == CLINFO_HUMAN ? 2 : 3);
 	const char * const *ptstr = (output_mode == CLINFO_HUMAN ?
 		partition_type_str : partition_type_raw_str);
 
+	set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
+
 	GET_VAL_ARRAY;
 
 	szval = 0;
@@ -1194,10 +1242,7 @@ int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, cons
 			int str_idx = -1;
 
 			/* add separator for values past the first */
-			if (szval > 0) {
-				strcpy(strbuf + szval, sep);
-				szval += sepsz;
-			}
+			add_separator(&szval);
 
 			switch (val[cursor]) {
 			case 0: str_idx = 1; break;
@@ -1206,7 +1251,7 @@ int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, cons
 			case CL_DEVICE_PARTITION_BY_AFFINITY_DOMAIN_EXT: str_idx = 4; break;
 			case CL_DEVICE_PARTITION_BY_NAMES_EXT: str_idx = 5; break;
 			default:
-				szval += sprintf(strbuf + szval, "by <unknown> (0x%" PRIX64 ")", val[cursor]);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "by <unknown> (0x%" PRIX64 ")", val[cursor]);
 				break;
 			}
 			if (str_idx > 0) {
@@ -1215,13 +1260,18 @@ int device_info_partition_types_ext(cl_device_id dev, cl_device_info param, cons
 				strncpy(strbuf + szval, ptstr[str_idx], slen);
 				szval += slen;
 			}
+			if (szval >= bufsz) {
+				trunc_strbuf();
+				break;
+			}
 		}
 		if (szval == 0) {
 			slen = strlen(ptstr[0]);
-			strcpy(strbuf, ptstr[0]);
+			memcpy(strbuf, ptstr[0], slen);
 			szval += slen;
 		}
-		strbuf[szval] = '\0';
+		if (szval < bufsz)
+			strbuf[szval] = '\0';
 	}
 
 	show_strbuf(pname, 0);
@@ -1244,21 +1294,18 @@ int device_info_partition_affinities(cl_device_id dev, cl_device_info param, con
 		 */
 		size_t szval = 0;
 		cl_uint i = 0;
-		const char *sep = (output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
-		size_t sepsz = (output_mode == CLINFO_HUMAN ? 2 : 3);
 		const char * const *affstr = (output_mode == CLINFO_HUMAN ?
 			affinity_domain_str : affinity_domain_raw_str);
+		set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
 		for (i = 0; i < affinity_domain_count; ++i) {
 			cl_device_affinity_domain cur = (cl_device_affinity_domain)(1) << i;
 			if (val & cur) {
 				/* match: add separator if not first match */
-				if (szval > 0) {
-					strcpy(strbuf + szval, sep);
-					szval += sepsz;
-				}
-				strcpy(strbuf + szval, affstr[i]);
-				szval += strlen(affstr[i]);
+				add_separator(&szval);
+				szval += bufcpy(szval, affstr[i]);
 			}
+			if (szval >= bufsz)
+				break;
 		}
 	}
 	if (val || had_error)
@@ -1271,12 +1318,11 @@ int device_info_partition_affinities_ext(cl_device_id dev, cl_device_info param,
 {
 	size_t numval = 0, szval = 0, cursor = 0, slen = 0;
 	cl_device_partition_property_ext *val = NULL;
-
-	const char *sep = (output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
-	size_t sepsz = (output_mode == CLINFO_HUMAN ? 2 : 3);
 	const char * const *ptstr = (output_mode == CLINFO_HUMAN ?
 		affinity_domain_ext_str : affinity_domain_raw_ext_str);
 
+	set_separator(output_mode == CLINFO_HUMAN ? comma_str : vbar_str);
+
 	GET_VAL_ARRAY;
 
 	szval = 0;
@@ -1285,10 +1331,7 @@ int device_info_partition_affinities_ext(cl_device_id dev, cl_device_info param,
 			int str_idx = -1;
 
 			/* add separator for values past the first */
-			if (szval > 0) {
-				strcpy(strbuf + szval, sep);
-				szval += sepsz;
-			}
+			add_separator(&szval);
 
 			switch (val[cursor]) {
 			case CL_AFFINITY_DOMAIN_NUMA_EXT: str_idx = 0; break;
@@ -1298,7 +1341,7 @@ int device_info_partition_affinities_ext(cl_device_id dev, cl_device_info param,
 			case CL_AFFINITY_DOMAIN_L1_CACHE_EXT: str_idx = 4; break;
 			case CL_AFFINITY_DOMAIN_NEXT_FISSIONABLE_EXT: str_idx = 5; break;
 			default:
-				szval += sprintf(strbuf + szval, "<unknown> (0x%" PRIX64 ")", val[cursor]);
+				szval += snprintf(strbuf + szval, bufsz - szval - 1, "<unknown> (0x%" PRIX64 ")", val[cursor]);
 				break;
 			}
 			if (str_idx >= 0) {
@@ -1308,6 +1351,10 @@ int device_info_partition_affinities_ext(cl_device_id dev, cl_device_info param,
 				strncpy(strbuf + szval, str, slen);
 				szval += slen;
 			}
+			if (szval >= bufsz) {
+				trunc_strbuf();
+				break;
+			}
 		}
 		strbuf[szval] = '\0';
 	}
@@ -1364,9 +1411,9 @@ int device_info_fpconf(cl_device_id dev, cl_device_info param, const char *pname
 	if (!had_error) {
 		size_t szval = 0;
 		cl_uint i = 0;
-		const char *sep = vbar_str;
 		const char * const *fpstr = (output_mode == CLINFO_HUMAN ?
 			fp_conf_str : fp_conf_raw_str);
+		set_separator(vbar_str);
 		if (output_mode == CLINFO_HUMAN) {
 			const char *why = na;
 			switch (param) {
@@ -1396,8 +1443,7 @@ int device_info_fpconf(cl_device_id dev, cl_device_info param, const char *pname
 					szval += sprintf(strbuf + szval, "\n%s" I2_STR "%s",
 						line_pfx, fpstr[i], bool_str[!!(val & cur)]);
 				} else if (val & cur) {
-					if (szval > 0)
-						szval += bufcpy(szval, sep);
+					add_separator(&szval);
 					szval += bufcpy(szval, fpstr[i]);
 				}
 			}
@@ -1419,17 +1465,16 @@ int device_info_qprop(cl_device_id dev, cl_device_info param, const char *pname,
 	if (!had_error) {
 		size_t szval = 0;
 		cl_uint i = 0;
-		const char *sep = vbar_str;
 		const char * const *qpstr = (output_mode == CLINFO_HUMAN ?
 			queue_prop_str : queue_prop_raw_str);
+		set_separator(vbar_str);
 		for (i = 0; i < queue_prop_count; ++i) {
 			cl_command_queue_properties cur = (cl_command_queue_properties)(1) << i;
 			if (output_mode == CLINFO_HUMAN) {
 				szval += sprintf(strbuf + szval, "\n%s" I2_STR "%s",
 					line_pfx, qpstr[i], bool_str[!!(val & cur)]);
 			} else if (val & cur) {
-				if (szval > 0)
-					szval += bufcpy(szval, sep);
+				add_separator(&szval);
 				szval += bufcpy(szval, qpstr[i]);
 			}
 		}
@@ -1451,17 +1496,16 @@ int device_info_execap(cl_device_id dev, cl_device_info param, const char *pname
 	if (!had_error) {
 		size_t szval = 0;
 		cl_uint i = 0;
-		const char *sep = vbar_str;
 		const char * const *qpstr = (output_mode == CLINFO_HUMAN ?
 			execap_str : execap_raw_str);
+		set_separator(vbar_str);
 		for (i = 0; i < execap_count; ++i) {
 			cl_device_exec_capabilities cur = (cl_device_exec_capabilities)(1) << i;
 			if (output_mode == CLINFO_HUMAN) {
 				szval += sprintf(strbuf + szval, "\n%s" I2_STR "%s",
 					line_pfx, qpstr[i], bool_str[!!(val & cur)]);
 			} else if (val & cur) {
-				if (szval > 0)
-					szval += bufcpy(szval, sep);
+				add_separator(&szval);
 				szval += bufcpy(szval, qpstr[i]);
 			}
 		}
@@ -1506,9 +1550,9 @@ int device_info_svm_cap(cl_device_id dev, cl_device_info param, const char *pnam
 	if (!had_error) {
 		size_t szval = 0;
 		cl_uint i = 0;
-		const char *sep = vbar_str;
 		const char * const *scstr = (output_mode == CLINFO_HUMAN ?
 			svm_cap_str : svm_cap_raw_str);
+		set_separator(vbar_str);
 		if (output_mode == CLINFO_HUMAN) {
 			/* show 'why' it's being shown */
 			szval += sprintf(strbuf, "(%s%s%s)",
@@ -1522,8 +1566,7 @@ int device_info_svm_cap(cl_device_id dev, cl_device_info param, const char *pnam
 				szval += sprintf(strbuf + szval, "\n%s" I2_STR "%s",
 					line_pfx, scstr[i], bool_str[!!(val & cur)]);
 			} else if (val & cur) {
-				if (szval > 0)
-					szval += bufcpy(szval, sep);
+				add_separator(&szval);
 				szval += bufcpy(szval, scstr[i]);
 			}
 		}
@@ -1607,6 +1650,7 @@ struct device_info_traits dinfo_traits[] = {
 	{ CLINFO_BOTH, DINFO(CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE, "Preferred work group size multiple", wg), NULL },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_WARP_SIZE_NV, "Warp size (NV)", int), dev_has_nv },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_WAVEFRONT_WIDTH_AMD, "Wavefront width (AMD)", int), dev_is_gpu_amd },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_MAX_NUM_SUB_GROUPS, "Max sub-groups per work group", int), dev_is_21 },
 
 	/* Preferred/native vector widths: header is only presented in HUMAN case, that also pairs
 	 * PREFERRED and NATIVE in a single line */
@@ -1736,6 +1780,7 @@ struct device_info_traits dinfo_traits[] = {
 
 	/* Kernel execution capabilities */
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_EXECUTION_CAPABILITIES, "Execution capabilities", execap), NULL },
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_SUB_GROUP_INDEPENDENT_FORWARD_PROGRESS, INDENT "Sub-group independent forward progress", bool), dev_is_21 },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_THREAD_TRACE_SUPPORTED_AMD, INDENT "Thread trace supported (AMD)", bool), dev_is_gpu_amd },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_KERNEL_EXEC_TIMEOUT_NV, INDENT "Kernel execution timeout (NV)", bool), dev_has_nv },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_GPU_OVERLAP_NV, "Concurrent copy and kernel execution (NV)", bool), dev_has_nv },
@@ -1744,6 +1789,7 @@ struct device_info_traits dinfo_traits[] = {
 	 * or maybe it depends on some other device property?
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_AVAILABLE_ASYNC_QUEUES_AMD, INDENT "Number of async queues (AMD)", int), dev_is_gpu_amd },
 	 */
+	{ CLINFO_BOTH, DINFO(CL_DEVICE_IL_VERSION, INDENT "IL version", str), dev_is_21, },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_SPIR_VERSIONS, INDENT "SPIR versions", str), dev_has_spir },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_PRINTF_BUFFER_SIZE, "printf() buffer size", mem), dev_is_12 },
 	{ CLINFO_BOTH, DINFO(CL_DEVICE_BUILT_IN_KERNELS, "Built-in kernels", str), dev_is_12 },
@@ -1768,15 +1814,15 @@ printDeviceInfo(const cl_device_id *device, cl_uint d,
 
 	char *extensions = NULL;
 
+	/* pointer to the traits for CL_DEVICE_EXTENSIONS */
+	const struct device_info_traits *extensions_traits = NULL;
+
 	struct device_info_checks chk;
 	memset(&chk, 0, sizeof(chk));
 	chk.dev_version = 10;
 
 	current_function = __func__;
 
-	/* pointer to the traits for CL_DEVICE_EXTENSIONS */
-	const struct device_info_traits *extensions_traits = NULL;
-
 	for (current_line = 0; current_line < ARRAY_SIZE(dinfo_traits); ++current_line) {
 
 		const struct device_info_traits *traits = dinfo_traits + current_line;
@@ -2204,7 +2250,7 @@ void checkNullCtx(cl_uint pidx, const cl_device_id *dev, const char *which)
 }
 
 /* check behavior of clCreateContextFromType() with NULL cl_context_properties */
-void checkNullCtxFromType()
+void checkNullCtxFromType(void)
 {
 	size_t t; /* type iterator */
 	size_t i; /* generic iterator */
@@ -2328,6 +2374,7 @@ void checkNullCtxFromType()
 		}
 		printf("%s%s\n", def, strbuf);
 	}
+	free(devs);
 }
 
 /* check the behavior of NULL platform in clGetDeviceIDs (see checkNullGetDevices)
@@ -2395,9 +2442,25 @@ typedef enum {
 	CL_ICDL_VENDOR=4,
 } cl_icdl_info;
 
-/* Function pointer to the ICD loader function */
+/* Function pointer to the ICD loader info function */
 cl_int (*clGetICDLoaderInfoOCLICD)(cl_icdl_info, size_t, void*, size_t*);
 
+/* We want to auto-detect the OpenCL version supported by the ICD loader.
+ * To do this, we will progressively find symbols introduced in new APIs,
+ * until a NULL symbol is found.
+ */
+
+struct icd_loader_test {
+	cl_uint version;
+	const char *symbol;
+} icd_loader_tests[] = {
+	{ 11, "clCreateSubBuffer" },
+	{ 12, "clCreateImage" },
+	{ 20, "clSVMAlloc" },
+	{ 21, "clGetHostTimer" },
+	{ 0, NULL }
+};
+
 int
 icdl_info_str(cl_icdl_info param, const char* pname)
 {
@@ -2421,6 +2484,8 @@ struct icdl_info_traits {
 	const char *pname; // "ICD loader *"
 };
 
+static const char * const oclicdl_pfx = "OCLICD";
+
 #define LINFO(symbol, name) { symbol, #symbol, "ICD loader " name }
 struct icdl_info_traits linfo_traits[] = {
 	LINFO(CL_ICDL_NAME, "Name"),
@@ -2429,12 +2494,32 @@ struct icdl_info_traits linfo_traits[] = {
 	LINFO(CL_ICDL_OCL_VERSION, "Profile")
 };
 
+/* GCC < 4.6 does not support the diagnostic push _inside_ the function,
+ * so we have to put it outside
+ */
+#if defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) < 46)
+#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
+#endif
 
-void oclIcdProps()
+void oclIcdProps(void)
 {
+	/* First of all, we try to auto-detect the supported ICD loader version */
+	int i = 0;
+
+	do {
+		struct icd_loader_test check = icd_loader_tests[i];
+		if (check.symbol == NULL)
+			break;
+		if (dlsym(RTLD_DEFAULT, check.symbol) == NULL)
+			break;
+		icdl_ocl_version_found = check.version;
+		++i;
+	} while (1);
+
+
 	/* We find the clGetICDLoaderInfoOCLICD extension address, and use it to query
 	 * the ICD loader properties. It should be noted however that
-	 * clGetExtensionFunctionAddress is marked * deprecated as of OpenCL 1.2, so
+	 * clGetExtensionFunctionAddress is marked deprecated as of OpenCL 1.2, so
 	 * to use it and compile cleanly we need disable the relevant warning.
 	 * It should be noted that in this specific case we cannot replace the
 	 * call to clGetExtensionFunctionAddress with a call to the superseding function
@@ -2445,7 +2530,7 @@ void oclIcdProps()
 #ifdef _MSC_VER
 #pragma warning(push)
 #pragma warning(disable : 4996)
-#else
+#elif defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) >= 46)
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wdeprecated-declarations"
 #endif
@@ -2454,14 +2539,23 @@ void oclIcdProps()
 
 #ifdef _MSC_VER
 #pragma warning(pop)
-#else
+#elif defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) >= 46)
 #pragma GCC diagnostic pop
 #endif
 
 	if (clGetICDLoaderInfoOCLICD != NULL) {
-		puts("\nICD loader properties");
+		/* TODO think of a sensible header in CLINFO_RAW */
+		if (output_mode != CLINFO_RAW)
+			puts("\nICD loader properties");
 		current_function = __func__;
 
+		if (output_mode == CLINFO_RAW) {
+			line_pfx_len = strlen(oclicdl_pfx) + 5;
+			REALLOC(line_pfx, line_pfx_len, "line prefix OCL ICD");
+			sprintf(strbuf, "[%s/*]", oclicdl_pfx);
+			sprintf(line_pfx, "%*s", -line_pfx_len, strbuf);
+		}
+
 		for (current_line = 0; current_line < ARRAY_SIZE(linfo_traits); ++current_line) {
 			const struct icdl_info_traits *traits = linfo_traits + current_line;
 			current_param = traits->sname;
@@ -2469,16 +2563,45 @@ void oclIcdProps()
 			had_error = icdl_info_str(traits->param,
 				output_mode == CLINFO_HUMAN ?
 				traits->pname : traits->sname);
+
+			if (!had_error && traits->param == CL_ICDL_OCL_VERSION) {
+				icdl_ocl_version = getOpenCLVersion(strbuf + 7);
+			}
+		}
+	}
+
+	if (output_mode == CLINFO_HUMAN) {
+		if (icdl_ocl_version &&
+			icdl_ocl_version != icdl_ocl_version_found) {
+			printf(	"\tNOTE:\tyour OpenCL library declares to support OpenCL %u.%u,\n"
+				"\t\tbut it seems to support up to OpenCL %u.%u %s.\n",
+				icdl_ocl_version / 10, icdl_ocl_version % 10,
+				icdl_ocl_version_found / 10, icdl_ocl_version_found % 10,
+				icdl_ocl_version_found < icdl_ocl_version  ?
+				"only" : "too");
+		}
+		if (icdl_ocl_version_found < max_plat_version) {
+			printf(	"\tNOTE:\tyour OpenCL library only supports OpenCL %u.%u,\n"
+				"\t\tbut some installed platforms support OpenCL %u.%u.\n"
+				"\t\tPrograms using %u.%u features may crash\n"
+				"\t\tor behave unexepectedly\n",
+				icdl_ocl_version_found / 10, icdl_ocl_version_found % 10,
+				max_plat_version / 10, max_plat_version % 10,
+				max_plat_version / 10, max_plat_version % 10);
 		}
 	}
 }
 
-void version()
+#if defined __GNUC__ && ((__GNUC__*10 + __GNUC_MINOR__) < 46)
+#pragma GCC diagnostic warning "-Wdeprecated-declarations"
+#endif
+
+void version(void)
 {
-	puts("clinfo version 2.0.15.04.28");
+	puts("clinfo version 2.1.16.01.12");
 }
 
-void usage()
+void usage(void)
 {
 	version();
 	puts("Display properties of all available OpenCL platforms and devices");
@@ -2563,7 +2686,8 @@ int main(int argc, char *argv[])
 		listPlatformsAndDevices(show_offline);
 	} else {
 		showDevices(show_offline);
-		checkNullBehavior();
+		if (output_mode != CLINFO_RAW)
+			checkNullBehavior();
 		oclIcdProps();
 	}
 
